;;; fdes.scm

;;	Copyright (C) 1995,1996,1997,1998,1999,2000,2002 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;; Return a boolean indicating whether @var{obj} is a port.
(define (port? obj)
  (or (input-port? obj)
      (output-port? obj)))

(define (file-position . args) (apply ftell args))
(define (file-set-position . args) (apply fseek args))

;; Move the underlying file descriptor for @var{port} to the integer
;; value @var{fdes} and sets its revealed count to one.  Any other ports
;; already using this descriptor will be automatically shifted to new
;; descriptors and their revealed counts reset to zero.
;; The return value is unspecified.
(define (move->fdes fd/port fd)
  (cond ((integer? fd/port)
	 (dup->fdes fd/port fd)
	 (close fd/port)
	 fd)
	(else
	 (primitive-move->fdes fd/port fd)
	 (set-port-revealed! fd/port 1)
	 fd/port)))

;; Decrement the revealed count for a port.
(define (release-port-handle port)
  (let ((revealed (port-revealed port)))
    (if (> revealed 0)
	(set-port-revealed! port (- revealed 1)))))

;; Return a new port using the new file descriptor.  @var{mode} supplies a
;; mode string for the port (@pxref{File Ports, open-file}).
(define (dup->port port/fd mode . maybe-fd)
  (let ((port (fdopen (apply dup->fdes port/fd maybe-fd)
		      mode)))
    (if (pair? maybe-fd)
	(set-port-revealed! port 1))
    port))

;; Return a new input port using the new file descriptor.
(define (dup->inport port/fd . maybe-fd)
  (apply dup->port port/fd "r" maybe-fd))

;; Return a new output port using the new file descriptor.
(define (dup->outport port/fd . maybe-fd)
  (apply dup->port port/fd "w" maybe-fd))

;; Return a new port if @var{port/fd} is a port, with the same mode as the
;; supplied port, otherwise returns an integer file descriptor.
(define (dup port/fd . maybe-fd)
  (if (integer? port/fd)
      (apply dup->fdes port/fd maybe-fd)
      (apply dup->port port/fd (port-mode port/fd) maybe-fd)))

;; Return a new port which is opened on a duplicate of the file
;; descriptor underlying @var{port}, with mode string @var{modes}
;; as for @ref{File Ports, open-file}.  The two ports share a file
;; position and file status flags.
;;
;; Unexpected behaviour can result if both ports are subsequently used
;; and the original and/or duplicate ports are buffered.  The mode string
;; can include @code{0} to obtain an unbuffered duplicate port.
;;
;; This procedure is equivalent to @code{(dup->port @var{port} @var{modes})}.
(define (duplicate-port port modes)
  (dup->port port modes))

;; Return an existing input port which has @var{fdes} as its underlying file
;; descriptor, if one exists, and increments its revealed count.
;; Otherwise, returns a new input port with a revealed count of 1.
(define (fdes->inport fdes)
  (let loop ((rest-ports (fdes->ports fdes)))
    (cond ((null? rest-ports)
	   (let ((result (fdopen fdes "r")))
	     (set-port-revealed! result 1)
	     result))
	  ((input-port? (car rest-ports))
	   (set-port-revealed! (car rest-ports)
			       (+ (port-revealed (car rest-ports)) 1))
	   (car rest-ports))
	  (else
	   (loop (cdr rest-ports))))))

;; Return an existing output port which has @var{fdes} as its underlying file
;; descriptor, if one exists, and increments its revealed count.
;; Otherwise, returns a new output port with a revealed count of 1.
(define (fdes->outport fdes)
  (let loop ((rest-ports (fdes->ports fdes)))
    (cond ((null? rest-ports)
	   (let ((result (fdopen fdes "w")))
	     (set-port-revealed! result 1)
	     result))
	  ((output-port? (car rest-ports))
	   (set-port-revealed! (car rest-ports)
			       (+ (port-revealed (car rest-ports)) 1))
	   (car rest-ports))
	  (else
	   (loop (cdr rest-ports))))))

;; Return the integer file descriptor underlying @var{port}.  As a
;; side effect the revealed count of @var{port} is incremented.
(define (port->fdes port)
  (set-port-revealed! port (+ (port-revealed port) 1))
  (fileno port))

;;; fdes.scm ends here
