;;; structs-records.scm

;;	Copyright (C) 1995,1996,1997,1998,1999,2000,2002 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; {Structs}

(define (struct-layout s)
  (struct-ref (struct-vtable s) vtable-index-layout))


;;; {Records}
;;;

;; Printing records: by default, records are printed as
;;
;;   #<type-name field1: val1 field2: val2 ...>
;;
;; You can change that by giving a custom printing function to
;; MAKE-RECORD-TYPE (after the list of field symbols).  This function
;; will be called like
;;
;;   (<printer> object port)
;;
;; It should print OBJECT to PORT.

(define (inherit-print-state old-port new-port)
  (if (get-print-state old-port)
      (port-with-print-state new-port (get-print-state old-port))
      new-port))

;; 0: type-name, 1: fields
(define record-type-vtable
  (make-vtable-vtable "prpr" 0
		      (lambda (s p)
			(cond ((eq? s record-type-vtable)
			       (display "#<record-type-vtable>" p))
			      (else
			       (display "#<record-type " p)
			       (display (record-type-name s) p)
			       (display ">" p))))))

(define (record-type? obj)
  (and (struct? obj) (eq? record-type-vtable (struct-vtable obj))))

;; Return a @dfn{record-type descriptor}, a value representing a new data
;; type disjoint from all others.  The @var{type-name} argument must be a
;; string, but is only used for debugging purposes (such as the printed
;; representation of a record of the new type).  The @var{field-names}
;; argument is a list of symbols naming the @dfn{fields} of a record of the
;; new type.  It is an error if the list contains any duplicates.  It is
;; unspecified how record-type descriptors are represented.  Optional third
;; argument @var{printer} is a procedure that takes two args, a record of
;; the newly defined type and an output port.  The default printer displays
;; a record along the lines of @code{#<TYPE-NAME FIELD1:VAL1 ...>}.
;;
;;-sig: (type-name fields [printer])
;;
(define (make-record-type type-name fields . opt)
  (let ((printer-fn (and (pair? opt) (car opt))))
    (let ((struct (make-struct record-type-vtable 0
			       (make-struct-layout
				(apply symbol-append
				       (map (lambda (f) "pw") fields)))
			       (or printer-fn
				   (lambda (s p)
				     (display "#<" p)
				     (display type-name p)
				     (let loop ((fields fields)
						(off 0))
				       (cond
					((not (null? fields))
					 (display " " p)
					 (display (car fields) p)
					 (display ": " p)
					 (display (struct-ref s off) p)
					 (loop (cdr fields) (+ 1 off)))))
				     (display ">" p)))
			       type-name
			       (copy-tree fields))))
      ;; Temporary solution: Associate a name to the record type descriptor
      ;; so that the object system can create a wrapper class for it.
      (set-struct-vtable-name! struct (if (symbol? type-name)
					  type-name
					  (string->symbol type-name)))
      struct)))

;; Return the type-name associated with the type represented by rtd.  The
;; returned value is @code{eqv?} to the @var{type-name} argument given in
;; the call to @code{make-record-type} that created the type represented by
;; @var{rtd}.
(define (record-type-name obj)
  (if (record-type? obj)
      (struct-ref obj vtable-offset-user)
      (error 'not-a-record-type obj)))

;; Return a list of the symbols naming the fields in members of the type
;; represented by @var{rtd}.  The returned value is @code{equal?} to the
;; field-names argument given in the call to @code{make-record-type} that
;; created the type represented by @var{rtd}.
(define (record-type-fields obj)
  (if (record-type? obj)
      (struct-ref obj (+ 1 vtable-offset-user))
      (error 'not-a-record-type obj)))

;; Return a procedure for constructing new members of the type represented
;; by @var{rtd}.  The returned procedure accepts exactly as many arguments
;; as there are symbols in the given list, @var{field-names}; these are
;; used, in order, as the initial values of those fields in a new record,
;; which is returned by the constructor procedure.  The values of any
;; fields not named in that list are unspecified.  The @var{field-names}
;; argument defaults to the list of field names in the call to
;; @code{make-record-type} that created the type represented by @var{rtd};
;; if the @var{field-names} argument is provided, it is an error if it
;; contains any duplicates or any symbols not in the default list.
;;
;;-sig: (rtd [field-names])
;;
(define (record-constructor rtd . opt)
  (let ((field-names (if (pair? opt) (car opt) (record-type-fields rtd))))
    (eval `(lambda ,field-names
	     (make-struct ',rtd 0 ,@(map (lambda (f)
					  (if (memq f field-names)
					      f
					      #f))
					(record-type-fields rtd)))))))

;; Return a procedure for testing membership in the type represented by
;; @var{rtd}.  The returned procedure accepts exactly one argument and
;; returns a true value if the argument is a member of the indicated record
;; type; it returns a false value otherwise.
(define (record-predicate rtd)
  (lambda (obj) (and (struct? obj) (eq? rtd (struct-vtable obj)))))

;; Return a procedure for reading the value of a particular field of a
;; member of the type represented by @var{rtd}.  The returned procedure
;; accepts exactly one argument which must be a record of the appropriate
;; type; it returns the current value of the field named by the symbol
;; @var{field-name} in that record.  The symbol @var{field-name} must be a
;; member of the list of field-names in the call to @code{make-record-type}
;; that created the type represented by @var{rtd}.
(define (record-accessor rtd field-name)
  (let* ((pos (list-index (record-type-fields rtd) field-name)))
    (if (not pos)
	(error 'no-such-field field-name))
    (eval `(lambda (obj)
	     (and (eq? ',rtd (record-type-descriptor obj))
		  (struct-ref obj ,pos))))))

;; Return a procedure for writing the value of a particular field of a
;; member of the type represented by @var{rtd}.  The returned procedure
;; accepts exactly two arguments: first, a record of the appropriate type,
;; and second, an arbitrary Scheme value; it modifies the field named by
;; the symbol @var{field-name} in that record to contain the given value.
;; The returned value of the modifier procedure is unspecified.  The symbol
;; @var{field-name} must be a member of the list of field-names in the call
;; to @code{make-record-type} that created the type represented by
;; @var{rtd}.
(define (record-modifier rtd field-name)
  (let* ((pos (list-index (record-type-fields rtd) field-name)))
    (if (not pos)
	(error 'no-such-field field-name))
    (eval `(lambda (obj val)
	     (and (eq? ',rtd (record-type-descriptor obj))
		  (struct-set! obj ,pos val))))))

;; Return @code{#t} if @var{obj} is a record of any type and @code{#f}
;; otherwise.
;;
;; Note that @code{record?} may be true of any Scheme value; there is no
;; promise that records are disjoint with other Scheme types.
(define (record? obj)
  (and (struct? obj) (record-type? (struct-vtable obj))))

;; Return a record-type descriptor representing the type of the given
;; record.  That is, for example, if the returned descriptor were passed to
;; @code{record-predicate}, the resulting predicate would return a true
;; value when passed the given record.  Note that it is not necessarily the
;; case that the returned descriptor is the one that was passed to
;; @code{record-constructor} in the call that created the constructor
;; procedure that created the given record.
(define (record-type-descriptor obj)
  (if (struct? obj)
      (struct-vtable obj)
      (error 'not-a-record obj)))

(provide 'record)

;;; structs-records.scm ends here
