;;; regex.scm

;;	Copyright (C) 1997, 1999, 2002 Free Software Foundation, Inc.
;;
;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.


;;;; POSIX regex support functions.

(define-module (ice-9 regex))

;;; FIXME:
;;;   It is not clear what should happen if a `match' function
;;;   is passed a `match number' which is out of bounds for the
;;;   regexp match: return #f, or throw an error?  These routines
;;;   throw an out-of-range error.

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; These procedures are not defined in SCSH, but I found them useful.

;; Return the number of parenthesized subexpressions from @var{match}.
;; Note that the entire regular expression match itself counts as a
;; subexpression, and failed submatches are included in the count.
(define-public (match:count match)
  (- (vector-length match) 1))

;; Return the original @var{target} string.
(define-public (match:string match)
  (vector-ref match 0))

;; Return the unmatched portion of @var{target} preceding the regexp match.
(define-public (match:prefix match)
  (make-shared-substring (match:string match)
			 0
			 (match:start match 0)))

;; Return the unmatched portion of @var{target} following the regexp match.
(define-public (match:suffix match)
  (make-shared-substring (match:string match)
			 (match:end match 0)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;; SCSH compatibility routines.

;; Return @code{#t} if @var{obj} is a match structure returned by a
;; previous call to @code{regexp-exec}, or @code{#f} otherwise.
(define-public (regexp-match? match)
  (and (vector? match)
       (string? (vector-ref match 0))
       (let loop ((i 1))
	 (cond ((>= i (vector-length match)) #t)
	       ((and (pair? (vector-ref match i))
		     (integer? (car (vector-ref match i)))
		     (integer? (cdr (vector-ref match i))))
		(loop (+ 1 i)))
	       (else #f)))))

;; Quote each special character found in @var{str} with a backslash, and
;; return the resulting string.
(define-public (regexp-quote regexp)
  (call-with-output-string
   (lambda (p)
     (let loop ((i 0))
       (and (< i (string-length regexp))
	    (begin
	      (case (string-ref regexp i)
		((#\* #\. #\( #\) #\+ #\? #\\ #\^ #\$ #\{ #\})
		 (write-char #\\ p)))
	      (write-char (string-ref regexp i) p)
	      (loop (1+ i))))))))

;; Return the starting position of submatch number @var{n}.
;;-sig: (match [n])
(define-public (match:start match . args)
  (let* ((matchnum (if (pair? args)
		       (+ 1 (car args))
		       1))
	 (start (car (vector-ref match matchnum))))
    (if (= start -1) #f start)))

;; Return the ending position of submatch number @var{n}.
;;sig: (match [n])
(define-public (match:end match . args)
  (let* ((matchnum (if (pair? args)
		       (+ 1 (car args))
		       1))
	 (end (cdr (vector-ref match matchnum))))
    (if (= end -1) #f end)))

;; Return the portion of @var{target} matched by subexpression number
;; @var{n}.  Submatch 0 (the default) represents the entire regexp match.
;; If the regular expression as a whole matched, but the subexpression
;; number @var{n} did not match, return @code{#f}.
;;
;;-sig: (match [n])
;;
(define-public (match:substring match . args)
  (let* ((matchnum (if (pair? args)
		       (car args)
		       0))
	 (start (match:start match matchnum))
	 (end   (match:end match matchnum)))
    (and start end (make-shared-substring (match:string match)
					  start
					  end))))

;; Compile the string @var{pattern} into a regular expression and compare
;; it with @var{str}.  The optional numeric argument @var{start} specifies
;; the position of @var{str} at which to begin matching.
;;
;; @code{string-match} returns a @dfn{match structure} which
;; describes what, if anything, was matched by the regular
;; expression.  @xref{Match Structures}.  If @var{str} does not match
;; @var{pattern} at all, @code{string-match} returns @code{#f}.
;;
;;-sig: (pattern str [start])
;;
(define-public (string-match pattern str . args)
  (let ((rx (make-regexp pattern))
	(start (if (pair? args) (car args) 0)))
    (regexp-exec rx str start)))

;; Write to the output port @var{port} selected contents of the match
;; structure @var{match}.  Each @var{item} specifies what should be
;; written, and may be one of the following arguments:
;;
;; @itemize @bullet
;; @item
;; A string.  String arguments are written out verbatim.
;;
;; @item
;; An integer.  The submatch with that number is written.
;;
;; @item
;; The symbol @samp{pre}.  The portion of the matched string preceding
;; the regexp match is written.
;;
;; @item
;; The symbol @samp{post}.  The portion of the matched string following
;; the regexp match is written.
;; @end itemize
;;
;; @var{port} may be @code{#f}, in which case nothing is written; instead,
;; @code{regexp-substitute} constructs a string from the specified
;; @var{item}s and returns that.
(define-public (regexp-substitute port match . items)
  ;; If `port' is #f, send output to a string.
  (if (not port)
      (call-with-output-string
       (lambda (p)
	 (apply regexp-substitute p match items)))

      ;; Otherwise, process each substitution argument in `items'.
      (for-each (lambda (obj)
		  (cond ((string? obj)   (display obj port))
			((integer? obj)  (display (match:substring match obj) port))
			((eq? 'pre obj)  (display (match:prefix match) port))
			((eq? 'post obj) (display (match:suffix match) port))
			(else (error 'wrong-type-arg obj))))
		items)))

;;; If we call fold-matches, below, with a regexp that can match the
;;; empty string, it's not obvious what "all the matches" means.  How
;;; many empty strings are there in the string "a"?  Our answer:
;;;
;;; 	This function applies PROC to every non-overlapping, maximal
;;;     match of REGEXP in STRING.
;;;
;;; "non-overlapping": There are two non-overlapping matches of "" in
;;; "a" --- one before the `a', and one after.  There are three
;;; non-overlapping matches of "q|x*" in "aqb": the empty strings
;;; before `a' and after `b', and `q'.  The two empty strings before
;;; and after `q' don't count, because they overlap with the match of
;;; "q".
;;;
;;; "maximal": There are three distinct maximal matches of "x*" in
;;; "axxxb": one before the `a', one covering `xxx', and one after the
;;; `b'.  Around or within `xxx', only the match covering all three
;;; x's counts, because the rest are not maximal.

(define-public (fold-matches regexp string init proc . flags)
  (let ((regexp (if (regexp? regexp) regexp (make-regexp regexp)))
	(flags (if (null? flags) 0 flags)))
    (let loop ((start 0)
	       (value init)
	       (abuts #f))		; True if start abuts a previous match.
      (let ((m (if (> start (string-length string)) #f
		   (regexp-exec regexp string start flags))))
	(cond
	 ((not m) value)
	 ((and (= (match:start m) (match:end m)) abuts)
	  ;; We matched an empty string, but that would overlap the
	  ;; match immediately before.  Try again at a position
	  ;; further to the right.
	  (loop (+ start 1) value #f))
	 (else
	  (loop (match:end m) (proc m value) #t)))))))

(define-public (list-matches regexp string . flags)
  (reverse! (apply fold-matches regexp string '() cons flags)))

;; Similar to @code{regexp-substitute}, but can be used to perform global
;; substitutions on @var{str}.  Instead of taking a match structure as an
;; argument, @code{regexp-substitute/global} takes two string arguments: a
;; @var{regexp} string describing a regular expression, and a @var{target}
;; string which should be matched against this regular expression.
;;
;; Each @var{item} behaves as in @var{regexp-substitute}, with the
;; following exceptions:
;;
;; @itemize @bullet
;; @item
;; A function may be supplied.  When this function is called, it will be
;; passed one argument: a match structure for a given regular expression
;; match.  It should return a string to be written out to @var{port}.
;;
;; @item
;; The @samp{post} symbol causes @code{regexp-substitute/global} to recurse
;; on the unmatched portion of @var{str}.  This @emph{must} be supplied in
;; order to perform global search-and-replace on @var{str}; if it is not
;; present among the @var{item}s, then @code{regexp-substitute/global} will
;; return after processing a single match.
;; @end itemize
(define-public (regexp-substitute/global port regexp string . items)

  ;; If `port' is #f, send output to a string.
  (if (not port)
      (call-with-output-string
       (lambda (p)
	 (apply regexp-substitute/global p regexp string items)))

      ;; Walk the set of non-overlapping, maximal matches.
      (let next-match ((matches (list-matches regexp string))
		       (start 0))
	(if (null? matches)
	    (display (make-shared-substring string start) port)
	    (let ((m (car matches)))

	      ;; Process all of the items for this match.  Don't use
	      ;; for-each, because we need to make sure 'post at the
	      ;; end of the item list is a tail call.
	      (let next-item ((items items))

		(define (do-item item)
		  (cond
		   ((string? item)    (display item port))
		   ((integer? item)   (display (match:substring m item) port))
		   ((procedure? item) (display (item m) port))
		   ((eq? item 'pre)
		    (display
		     (make-shared-substring string start (match:start m))
		     port))
		   ((eq? item 'post)
		    (next-match (cdr matches) (match:end m)))
		   (else (error 'wrong-type-arg item))))

		(if (pair? items)
		    (if (null? (cdr items))
			(do-item (car items)) ; This is a tail call.
			(begin
			  (do-item (car items)) ; This is not.
			  (next-item (cdr items)))))))))))

;;; regex.scm ends here
