;;; lr0-debug.scm

;;	Copyright (C) 2002 Free Software Foundation, Inc.
;;
;; This file is part of GUILE
;;
;; GUILE is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; GUILE is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; Author: Thien-Thi Nguyen <ttn@gnu.org>

;;; Commentary:

;; In lr0.scm there is a comment:
;;
;;   Some day, grammar debugging tools will be added to this library.
;;
;; Well, that day is here!  This module is a drop-in replacement for
;; module (lang lr0) with some additional debugging features.
;;
;; TODO: Document better.

;;; Code:

(define-module (lang lr0-debug)
  :use-module (ice-9 common-list)
  :use-module (lang debug-common)
  :use-module (lang grammar)
  :use-module ((lang lr0) :renamer (symbol-prefix-proc 'ACTUAL:))
  :export (;; mimic exports from module (lang lr0)
           item-set-start-kernel
           item-set-successor-kernel
           item-set-reductions
           item-set-closure
           grammar-item-table
           ;; "native" (defined here)
           VALIDATE:production
           VALIDATE:item
           VALIDATE:item-set
           grammar-with-nicknames
           display-grammar-item-table
           display-item-set
           display-item-set-closure))

;; intimacy w/ module (lang lr0)

(define *lr0* (resolve-module '(lang lr0)))

(defmacro intimate-lr0 (name)
  `(define ,name (module-ref *lr0* ',name)))

(intimate-lr0 item-production)
(intimate-lr0 item-position)
(intimate-lr0 item-index)
(intimate-lr0 item-set)

;; type invariance validation

(define (VALIDATE:production obj)
  (or (list? obj)
      (error "supposed production not a list:" obj))
  (or (< 2 (length obj))
      (error "supposed production has length not >= 2:" obj))
  (or (symbol? (car obj))
      (error "supposed production car not a symbol:" (car obj)))
  (or (every symbol? (cadr obj))
      (error "supposed production cadr not a list of symbols:" (cadr obj))))

(define (VALIDATE:item obj)
  (or (list? obj)
      (error "supposed item not a list:" obj))
  ;; car is the production: (SYMBOL (SYM SYM SYM) . REDUX)
  (VALIDATE:production (car obj))
  ;; cdr is some subset of the production expansion
  (let ((expansion (cadr (car obj))))
    (or (some (lambda (n)
                (eq? (list-tail expansion n)
                     (cdr obj)))
              (iota (1+ (length expansion))))
        (error "supposed item cdr:" (cdr obj) "not a subset of the expansion:"
               expansion)))
  obj)

(define (VALIDATE:item-set obj)
  (or (list? obj)
      (error "supposed item-set not a list: " obj))
  (for-each VALIDATE:item obj)
  obj)

;; hashing for display

(define (grammar-with-nicknames grammar)
  (let ((ht (make-hash-table 31)))
    (for-each (lambda (p)
                (VALIDATE:production p)
                (let ((sym (production-symbol p)))
                  (nick! p (format #f "~A[~A]"
                                   sym
                                   (hashq-set! ht sym
                                               (1+ (or (hashq-ref ht sym)
                                                       -1)))))))
              grammar))
  grammar)

;; display

(define (display-item item)
  (format #t "~A: ~A | ~A\n"
          (item-index item)
          (nick (item-production item))
          (item-position item)))

(define (display-grammar-item-table table)
  (do ((i 0 (1+ i)))
      ((= i (vector-length table)))
    (display-item (vector-ref table i))))

(define (display-item-set is)
  (for-each display-item is))

(define (display-item-set-closure isc)
  (for-each display-item isc))

;; poke debugging vars

(module-define! *lr0* 'item-closure-debug-pk
                (lambda (item caar-stack actual-x)
                  (display-item item)
                  (format #t "caar-stack: ")
                  (display-item caar-stack)
                  (for-each (lambda (item)
                              (format #t "--actual-x: ")
                              (display-item item))
                            actual-x)))

;; wrap the actual

(define item-set-start-kernel     ACTUAL:item-set-start-kernel)
(define item-set-successor-kernel ACTUAL:item-set-successor-kernel)
(define item-set-reductions       ACTUAL:item-set-reductions)
(define item-set-closure          ACTUAL:item-set-closure)
(define grammar-item-table        ACTUAL:grammar-item-table)

;;; lr0-debug.scm ends here
