;;; lr0.scm

;;	Copyright (C) 1996,2002 Free Software Foundation, Inc.
;;
;; This file is part of GUILE
;;
;; GUILE is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; GUILE is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; Commentary:

;; This module provides tools for LR(0) DFA construction:
;;  Debugging Support
;;  Item
;;  Item Functions
;;  Item Set
;;  Superstate Functions
;;
;; A grammar (see grammar.scm) implies a set of "items" which can be
;; interpreted as states in an NFA.  An item corresponds to a choice of one
;; production from the grammar, and a position within that production.
;;
;; For example, if a production is:
;;
;;  A -> B C D                  aka     (a (b c d) . <reduction rule>)
;;
;; Then the items are:
;;
;;  A -> * B C D                aka     ((a (b c d) . <reduction rule>) b c d)
;;  A -> B * C D                aka     ((a (b c d) . <reduction rule>) c d)
;;  A -> B C * D                aka     ((a (b c d) . <reduction rule>) d)
;;  A -> B C D *                aka     ((a (b c d) . <reduction rule>))
;;
;; An empty production has exactly one item:
;;
;;  A -> *                      aka     ((a () . <reduction rule>))
;;
;; If an item has nothing to the left of the star, it is called an "initial item".
;; If an item has nothing to the right of the star, it is called a "final item".
;;
;; An item says what we think we are parsing (an A, in the above examples)
;; and what we expect the next bit of input to be (for example, in
;; the item A -> * B C D, we next expect to parse a B).
;;
;; These items are states in an NFA, the LR(0) NFA, with edges defined
;; between them.
;;
;; The first item in the grammar is conventionally the start state, although
;; any initial item can be used for this purpose.
;;
;; Final states are items with no more expansion to the right of the *.
;; A final state implies that an entire expansion has been seen and now
;; a reduction of that expansion is possible.   Final states have no
;; outward edges.
;;
;; An NFA state (aka an item) has two kinds of edges corresponding to whether
;; the symbol to the right of the star is a terminal or non-terminal.
;;
;; Terminal edges indicate that a particular type of input token is expected.
;; If the lookahead is of that type, it can be shifted onto the parse stack
;; whie crossing the terminal edge to a new state.
;;
;; Non-terminal edges indicate a point where a nested construct should be
;; parsed.  If in state I the symbol following the star is non-terminal B,
;; then there is an epsilon edge to all initial items associated with
;; productions of B.  In case one of those epsilon edges leads to a correct
;; parse, there is also an edge labeled B that can be used to shift the B
;; generated by a reduction.
;;
;; For example, in the state:
;;
;;      A -> w * B u
;;
;; it is legal to shift the non-terminal B (this is sometimes called a "goto"
;; move).  It is also legal to make an epsilon transition to states:
;;
;;      B -> * w
;;      B -> * u
;;      ...
;;
;; presuming the corresponding productions exist.  Informally: you're
;; allowed to absorb the next burst of input as a reduction to B,
;; therefore, you are also allowed at this point to begin working on
;; a parse of B.
;;
;; The rule applies transitively so that if A -> w * B u epsilon
;; transitions to B -> * C v, that in turn epsilon transitions
;; to C -> * z.
;;
;; To summarize:
;;
;; The lr(0) NFA has final states which specify reductions,
;; terminal transitions, which specify permissible input,
;; non-terminal transitions, which specify permissible feedback from
;; the parsing of nested constructs, and epsilon edges, which specify
;; points at which nested constructs may begin.
;;
;;
;; The above describes all the states, designates the start and final
;; states and defines the transition function of a non-deterministic
;; automata.  The epsilon edges associated with non-terminal transitions
;; prevent the automata from being deterministic.
;;
;; But the lr(1) parser requires a deterministic lr(0) automata
;; So the code in this file not only must construct the lr(0) NFA,
;; but it must convert that NFA into a DFA.
;;
;; To parse deterministically, we apply a few stern measures.
;; First, we combine all states accessible by epsilon transitions
;; into superstates.   Second, when components of a superstate disagree
;; about whether to shift some input or reduce because of it, we favor
;; shifting.   Third, we only permit reductions if we can guarantee
;; that after a series of reductions, the look-ahead token will be
;; shifted.   Fourth, if the components of a superstate disagree
;; about which reduction rule to apply, we pick whichever one we
;; discover first (clearly this could be improved). Fifth and finally,
;; if only some components of a superstate think the look-ahead
;; token is an error, we ignore them (if all components agree the look-ahead
;; is an error, then so do we).
;;
;; Its useful to prove that a particular grammar has the property that
;; components of an achievable superstate will never disagree about
;; what reduction rule to apply.  This can be proven automatically.
;; One convenient technique for lalr(1) grammars is to convert them
;; to yacc syntax and run a table generator on them.  Some day, grammar
;; debugging tools will be added to this library.

;;; Code:

(define-module (lang lr0)
  :use-module (lang grammar)
  :use-module (ice-9 common-list)
  :use-module (ice-9 hcons)
  :export (grammar-item-table
           item-set-start-kernel
           item-set-successor-kernel
           item-set-reductions
           item-set-closure))


;; {Debugging Support}
;;
;; These are variables that are set by (lang lr0-debug) when it is used as a
;; drop-in replacement for (lang lr0).  They are not exported, and their use
;; may change unexpectedly.  Do not rely on them.
;;

(define item-closure-debug-pk #f)


;; {Item}
;;

;; Return an item from GRAMMAR PRODUCTION at a certain position POS.
;; Note that POS mus be eq? to some tail of the expansion of PRODUCTION
;; and that PRODUCTION must not share any cons pairs with any other
;; grammar production.  The returned item is associated with GRAMMAR.
;;
(define (make-item grammar production pos)
  (hashq-cons (grammar-cache grammar) production pos))

;; Return the production corresponding to ITEM.
;;
(define (item-production item) (car item))

;; Return the position, a (possibly empty) list, corresponding to ITEM.
;; This is also known as its "right hand expansion".
;;
(define (item-position item) (cdr item))


;; {Item Functions}
;;

;; Return the item table for GRAMMAR, making it if necessary.
;;
;; Every grammar is given an item table: a vector of all items
;; associated with the grammar.  An item's position in this table
;; is called the item's "index".
;;
;; The extent of the table is tied to the extent of GRAMMAR, and the
;; extent of items to the table.  For a given production, and position,
;; make-item (see below) will always return eq? items from the
;; item-table of G.
;;
(define (grammar-item-table g)
  (or (hashq-ref (grammar-cache g) 'grammar-item-table)
      (let* ((p-items (lambda (p)
                        (let loop ((pos (production-expansion p))
                                   (so-far '()))
                          (if (null? pos)
                              (cons (make-item g p '()) so-far)
                              (loop (cdr pos)
                                    (cons (make-item g p pos) so-far))))))
             (table (list->vector (apply append! (map p-items g)))))
        ;; make back-pointers
        (do ((i 0 (1+ i)))
            ((= i (vector-length table)))
          (set-object-property! (vector-ref table i) 'item-index i))
        (hashq-set! (grammar-cache g) 'grammar-item-table table))))

;; Return item-index of GRAMMAR item ITEM.
;;
(define (item-index item)
  (or (object-property item 'item-index)
      (error (format #f "item does not have an item index: ~A" item))))


;; {Item Set}

;; Return an item set built from a GRAMMAR's ITEMS.
;; The item set has the same extent as GRAMMAR.
;;
;; Item sets are represented as a (possibly empty) list of items.  Sets
;; have the property that equal? sets are eq?, and that elements of each
;; set are ordered from highest-numbered index to lowest.
;;
(define (item-set g . items)
  (let* ((cache (grammar-cache g))
         (table (grammar-item-table g))
         (flags (make-string (vector-length table) #\000)))
    ;; Use flag vector for O(n+m), n: items count, m: table length.
    ;; Another way, more complex: (uniq (sort (map item-index items) <)).
    (for-each (lambda (item)
                (string-set! flags (item-index item) #\001))
              items)
    (let loop ((base 0)
               (set '()))
      (cond ((string-index flags #\001 base)
             => (lambda (n)
                  (loop (1+ n)
                        (hashq-cons cache (vector-ref table n) set))))
            (else set)))))

;; Return an item set built from a union of GRAMMAR's item sets SETS.
;;
(define (item-set-union g sets)
  (apply item-set g (apply append sets)))


;; {Superstate Functions}
;;
;; In the usual way, we can construct a DFA from an LR(0) NFA, combining
;; multiple NFA states accessible along epsilon paths into singular DFA
;; superstates.


;; Return the kernel of the starting superstate (an item-set of sorts).
;;
;; Only the initial items of start-symbol productions are included in
;; this set.  Compose with item-set-closure to also obtain all items
;; reachable by epsilon transitions.
;;
(define (item-set-start-kernel g)
  (map (lambda (production)
         (make-item g production (production-expansion production)))
       (symbol-productions g (grammar-start-symbol g))))

;; With item-set IS, shift symbol S; return the resulting (possibly
;; empty) item-set.
;;
;; Only the items which are direct successors of items in IS are present
;; in the successor kernel.  To also include all items reachable by
;; epsilon transitions, use item-set-closure.
;;
(define (item-set-successor-kernel g is symbol)
  (pick-mappings (lambda (item)
                   (let ((prod (item-production item))
                         (pos (item-position item)))
                     (and (not (null? pos))
                          (eq? symbol (car pos))
                          (make-item g prod (cdr pos)))))
                 is))

;; Return the (possibly empty) set of all productions that can be
;; reduced from IS, disregarding the context and look-ahead.
;;
(define (item-set-reductions g is)
  (pick-mappings (lambda (item)
                   (and (null? (item-position item))
                        (item-production item)))
                 is))

;; Compute the partial epsilon transition closure for a GRAMMAR's ITEM.
;; This is used by `item-set-closure', primarily.
;;
(define item-closure
  (let (;; A memo of closures already computed.
	;; [I'm still trying to understand why this is better than either an
        ;;  object property of the item, or using the grammar cache.  --ttn]
	(memo (make-weak-key-hash-table 1031)))

    (lambda (g it)

      ;; Return the (possibly empty) set of initial items for non-terminal S.
      ;;
      (define (symbol-item-set-kernel-members s)
        (map (lambda (production)
               (make-item g production (production-expansion production)))
             (symbol-productions g s)))

      (cond
       ((hashq-ref memo it))

       (else

        ;; This is the digraph algorithm from "Efficient Construction of
        ;; LALR(1) Lookahead Sets" by F. DeRemer and T. Pennello, in ACM
        ;; TOPLS Vol 4 No 4, October 1982.  They credit J. Eve and
        ;; R. Kurki Suonio, "On Computing the transitive closure of a
        ;; relation."  Acta Inf. 8 1977.
        ;;
        ;; It is used here to compute epsilon closures of individual items.
        ;;
        (let traverse ((x it)
                       (approximations
                        ;; alist of elements: (ITEM DEPTH . APPROX-X-CLOSURE)
                        '()))
          (let* ((depth (1+ (if (null? approximations)
                                0
                                (cadar approximations))))
                 (next-states (symbol-item-set-kernel-members
                               (let ((pos (item-position x)))
                                 ;; #f is never a valid non-terminal, so
                                 ;; pos '() => next-states '().
                                 (and (not (null? pos))
                                      (car pos)))))
                 (approx-x-closure (apply item-set g x next-states))
                 (new-approximations (acons x (cons depth approx-x-closure)
                                            approximations)))
            (let next-loop ((next-states next-states)
                            (current-approx new-approximations))
              (if (not (null? next-states))
                  (let* ((y (car next-states))
                         ;; can be #f
                         (x-approx (assq-ref current-approx x))
                         ;; can be #f / can never be '()
                         (y-approx (or (assq-ref current-approx y)
                                       (and=> (hashq-ref memo y)
                                              (lambda (v)
                                                (cons (car x-approx) v)))))
                         (better-approx (if y-approx
                                            current-approx
                                            (traverse y current-approx))))
                    (next-loop (cdr next-states) better-approx))
                  (if (not (= depth (cadr (assq x current-approx))))
                      current-approx
                      (let ((actual-x (apply item-set g
                                             (cddr (assq x current-approx)))))
                        (let unify-component-loop ((stack current-approx))
                          (or stack (error "stack #f"))
                          (or (null? stack)
                              (begin
                                (cond (item-closure-debug-pk
                                       => (lambda (proc)
                                            (proc it (caar stack) actual-x))))
                                ;; jam it
                                (hashq-set! memo (caar stack) actual-x)
                                (if (not (eq? x (caar stack)))
                                    (unify-component-loop (cdr stack))
                                    (cdr stack)))))))))))
        (hashq-ref memo it))))))

;; Return the epsilon transition closure of is.
;;
;; In the LR(0) NFA, if we are at the state:
;;
;;         A -> w * B v
;;
;; Then we are also an epsilon transition (not to be confused
;; with an empty reduction) away from also being in the items:
;;
;;         B -> * u
;;         B -> * r
;;         B -> * s
;;         ...
;;
;; So if a superstate contains A->w*Bv, then it should also contain
;; the inital states of B.  This rule applies transitively.
;;
;; If IS was returned by item-set-start-kernel.  then this procedure
;; returns the DFA start state.
;;
;; If IS was returned by item-set-successor-kernel, and the IS argument
;; to item-set-successor-kernel was a DFA state, then this procedure
;; returns a DFA state.
;;
(define (item-set-closure g is)
  (item-set-union g (map (lambda (i)
                           (item-closure g i))
                         is)))

;;; lr0.scm ends here
