;;; lr1.scm

;;	Copyright (C) 1996,2002 Free Software Foundation, Inc.
;;
;; This file is part of GUILE
;;
;; GUILE is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; GUILE is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; Commentary:

;; This module provides tools for LR(1) parser construction:
;;  Debugging Support
;;  Value and State Stacks
;;  Parser State
;;  Production Function
;;  Parser
;;
;; An LR(1) parser is an abstract machine with these parts:
;;
;;   lr0-dfa
;;   state-stack
;;   value-stack
;;   lexer
;;   parse-error
;;
;; The function "parse-error" is called if the stream returned by the lexer is
;; unparsable.
;;
;; The function "lexer" is an input stream that returns a new input token each
;; time it is called.  Every input token belongs to a category of input tokens
;; and the category is named by a "terminal symbol".  For example,
;; "define-module" might be the token, and "identifier" the corresponding
;; terminal symbol.
;;
;; The value stack is a stack of arbitrary values, constructed as the parse
;; proceeds.  For example, the value stack can be used to build up parse trees
;; of input.
;;
;; The state-stack is a stack of lr0-dfa states.
;;
;; The lr0-dfa is a deterministic finite automata with these parts:
;;
;; 	states 			-- a set of states
;; 	next(state, symbol)	-- a transition function
;; 	reductions(state)	-- see below
;;
;; If you aren't familiar with lr(1) parsing and are reading these comments to
;; become familiar with it, don't worry yet about the details of the lr0-dfa.
;; Just get an idea of what the basic moves are for the parser machine.  Then
;; later, when the lr0-dfa is explained in detail, you can understand it by
;; thinking of how the parser machine would behave for such a dfa.
;;
;; Parsing begins with a singleton state stack containing an arbitrarily
;; chosen "start state" from lr0-dfa.  The value stack also starts off with an
;; arbitrarily chosen value.
;;
;; The parser reads one token, called the look-ahead, and evaluates:
;;
;; 	shift-state = next (top (state-stack),
;; 			    terminal-symbol (look-ahead))
;;
;; Evaluation may lead to finding "shift-state", or it may lead to an error if
;; no appropriate transition is defined.
;;
;; If a transition is defined, then the parser pushes shift-state onto the
;; state-stack, look-ahead onto the value-stack and repeats the cycle, reading
;; a new look-ahead token.
;;
;; After building up the stacks in this way, the parser may reach a point at
;; which shift-state is not defined for the current look-ahead.  In this case,
;; the parser consults:
;;
;; 	reductions (top (state-stack))
;;
;; "reductions" returns a set of reduction procedures.  A reduction procedure
;; is a mapping from parser states to parser states.  That is, a reduction
;; procedure takes a state-stack and a value-stack as input, and returns a new
;; state-stack, new value-stack as output.
;;
;; Reduction procedures always work by popping 0 or more values from both the
;; state-stack and value-stack (popping an equal number from each) and then
;; pushing exactly one new value and one new state.  If, after popping values,
;; the top of the state stack is S, then a reduction procedure will push S'
;; which is some next(S,X) for some symbol X.  For a given reduction
;; procedure, X is always the same, but S' may vary depending on S.
;; Conventionally, X is some symbol which is not a member of the set of
;; terminal symbols.
;;
;; After S' has been pushed, it may or may not be possible to shift the
;; look-ahead.  If it is not possible to shift the look-ahead, it may or may
;; not be possible to perform further reductions after which it would be
;; possible to shift the look-ahead.
;;
;; The parser defined here never performs _any_ reduction without first
;; computing a complete sequence of reductions that will lead to the
;; look-ahead being shiftable.  If no such sequence can be found, an error is
;; reported.  If multiple reductions are possible at some point, but only one
;; of them leads to the look-ahead being shifted, that one will be chosen.
;;
;; After performing all the necessary reductions to permit shifting of the
;; look-ahead, the parser shifts the look-ahead, reads a new look-ahead token,
;; and restarts its cycle.

;;; Code:

(define-module (lang lr1)
  :use-module (guile-user)
  :use-module (lang lr0)
  :use-module (lang grammar)
  :use-module (ice-9 hcons)
  :export (parser-value-cons
           parser-value-car
           parser-value-cdr
           parser-value-cdr-ref
           parser-context-cons
           parser-context-car
           parser-context-cdr
           parser-context-cdr-ref
           make-parser-state
           parser-state-item-set
           parser-start-state
           parser-action
           lr1-parse))


;; {Debugging Support}
;;
;; These are variables that are set by (lang lr1-debug) when it is used as a
;; drop-in replacement for (lang lr1).  They are not exported, and their use
;; may change unexpectedly.  Do not rely on them.
;;

(define parser-context-cons-add-nickname #f)
(define parser-loop-pk #f)


;; {Value and State Stacks}

;; The value stack holds tokens as they are shifted,
;; and the return values of reduction functions as they
;; are used.
;;
(define parser-value-cons cons)
(define parser-value-car car)
(define parser-value-cdr cdr)
(define parser-value-cdr-ref list-tail)

;; The state stack holds lr(0) states as they are shifted.
;;
(define (parser-context-cons g state context)
  (cons ((or parser-context-cons-add-nickname identity) state)
        context))
(define parser-context-car car)
(define parser-context-cdr cdr)
(define parser-context-cdr-ref list-tail)


;; {Parser State}
;;
;; Parser state has the form: (ITEM-SET . ACTION-CACHE)
;; and is cached using the key: (ITEM-SET . parser-state)
;; where `parser-state' is a literal symbol.

(define (make-parser-state g is)
  (let* ((cache (grammar-cache g))
	 (key (hashq-cons cache is 'parser-state)))
    (or (hashq-ref cache key)
        (hashq-set! cache key (cons is '())))))

(define parser-state-item-set car)

(define parser-state-action-cache cdr)

;; The action-cache is an alist with elements: (TOKEN-TYPE . ACTION).

(define (parser-state-cache-action! state token-type action)
  (set-cdr! state (acons token-type action (parser-state-action-cache state)))
  action)


;; {Production Function}

;; Production functions are called on a reduce.  They invoke the reduction
;; functions specified in the grammar and then callback to the parser-loop.
;; A backtracking procedure is pushed onto the backtrack-stack.
;;
;; [NB: This is my undertanding so far, may not be complete! --ttn]

(define (production-function g production module)
  (let* ((cache (grammar-cache g))
         (key (hashq-cons cache production 'production-function)))
    (or (hashq-ref cache key)
        (hashq-set!
         cache key
         (let ((symbol (production-symbol production))
               (body (production-body production))
               (arity (length (production-expansion production))))
           (let* ((dollar-vars
                   (let recursive ((n arity))
                     (if (= n 0)
                         '()
                         (cons (symbol-append '$ (number->string n))
                               (recursive (+ -1 n))))))
                  (value-stack-mutator
                   (eval-in-module
                    `(lambda (value-stack)
                       (delay
                         (let* ((value-stack (force value-stack))
                                (stack-tail (list-tail value-stack ,arity))
                                (bindings (list-head value-stack ,arity)))
                           ;; This is properly `parser-value-cons' instead of
                           ;; simply `cons'.  However, its use necessitates
                           ;; access to the definition by the user module,
                           ;; which is wrong.
                           (cons
                            (apply (lambda (value-stack ,@dollar-vars)
                                     ,@body)
                                   stack-tail
                                   bindings)
                            stack-tail))))
                    module)))
             (lambda (remaining-reductions
                      parser-loop
                      state-stack
                      value-stack
                      backtrack-stack
                      token
                      lexer
                      parse-error)

               ;; A reduction function does three things before
               ;; continuing the parser loop...
               ;;
               (parser-loop

                ;; ... it shifts the non-terminal
                ;; generated by the reduction.
                ;;
                (let ((popped (parser-context-cdr-ref state-stack arity)
                              ;;
                              ;; This shift take place on the stack underneath
                              ;; the states being removed by the reduction.
                              ))

                  (parser-context-cons g
                                       ;; The action associated with a
                                       ;; just-reduced non-terminal symbol is
                                       ;; always a `(shift ,new-state)
                                       ;;
                                       (cadr (parser-action
                                              g (parser-context-car popped)
                                              symbol module))

                                       popped))

                ;; ... it (lazily) performs the body of
                ;; the associated grammar production to
                ;; compute a new value stack.
                ;;
                (value-stack-mutator value-stack)

                ;; ... it remembers how to back-off this
                ;; reduction in case it is discovered
                ;; that the next token can not be
                ;; shifted otherwise.
                ;;
                (cons (lambda (parser-loop
                               bogus-state-stack
                               bogus-value-stack
                               our-backtrack-stack
                               token
                               lexer
                               parse-error)
                        (if (not (null? remaining-reductions))
                            ;; If there are any alternative reductions to
                            ;; consider, try those first.
                            ;;
                            ((car remaining-reductions) (cdr remaining-reductions)
                             parser-loop
                             state-stack
                             value-stack
                             backtrack-stack
                             token
                             lexer
                             parse-error)

                            ;; When there are no more reductions to consider,
                            ;; try further backtracking.
                            ;;
                            ((car backtrack-stack) parser-loop
                             state-stack
                             value-stack
                             backtrack-stack
                             token
                             lexer
                             parse-error)))
                      backtrack-stack)
                token
                lexer
                parse-error))))))))


;; {Parser}

;; Return the initial parser state derived from GRAMMAR's
;; item-set-start-kernel.  The return value is cached.
;;
(define (parser-start-state g)
  (let ((cache (grammar-cache g)))
    (or (hashq-ref cache 'parser-start-state)
        (hashq-set! cache 'parser-start-state
                    (make-parser-state
                     g (item-set-closure g (item-set-start-kernel g)))))))

;; Given GRAMMAR's STATE and the current TOKEN-TYPE, determine the action to
;; take (shift, reduce, or error) and return the form:
;;   (ACTION [DATA])
;; Where DATA is next state (if shifting), a list of the production functions
;; (if reducing), or empty (if error).  The return value is cached.
;;
(define (parser-action g state token-type module)
  (or (assq-ref (parser-state-action-cache state) token-type)
      (parser-state-cache-action!
       state token-type
       (or
        ;; shift
        (let ((next (item-set-successor-kernel g (parser-state-item-set state)
                                               token-type)))
          (and (not (null? next))
               `(shift ,(make-parser-state g (item-set-closure g next)))))
        ;; reduce
        (let ((all-r (item-set-reductions g (parser-state-item-set state))))
          (and (not (null? all-r))
               `(reduce ,(map (lambda (r)
                                (production-function g r module))
                              all-r))))
        ;; error
        '(error)))))

;;--ttn-tmp (define ---count--- 0)

;; Parse GRAMMAR, using token-providing procedure LEXER and error-handling
;; procedure PARSER-ERROR, doing reductions in the environment of module
;; MODULE.
;;
;; [At this time, I don't know if this procedure returns.  So far, I've only
;;  tested with a throw in the reduction body. --ttn]
;;
(define (lr1-parse g lexer parse-error module)

  (let (;; The backtrack always contains at least one entry: the
	;; caller's error procedure.
	;;
	(base-backtrack-stack
	 (list (letrec ((bottom-backtrack
			 (lambda (loop states values backs token lexer err)
			   (err (lambda (states values token lexer parse-error)
				  (loop states values (list bottom-backtrack)
					token lexer parse-error))
				states
				values
				token
				lexer))))
		 bottom-backtrack))))

    (let parser-loop ((state-stack (parser-context-cons
                                    g (parser-start-state g) '()))
		      (value-stack (parser-value-cons
                                    (cons (grammar-start-symbol g) '()) '()))
		      (backtrack-stack base-backtrack-stack)
		      (token (lexer))
		      (lexer lexer)
		      (parse-error parse-error))

;;--ttn-tmp      (set! ---count--- (1+ ---count---))
;;--ttn-tmp      (cond ((> ---count--- 100)
;;--ttn-tmp             (format #t "ARTIFICIAL ABORT THROW!\n")
;;--ttn-tmp             (throw '%%YY 'abort #f)))

      ;;(pk 0 (length state-stack))

      (let ((action (parser-action g (parser-context-car state-stack)
                                   (car token) module)))

        (cond (parser-loop-pk
               => (lambda (proc)
                    (proc state-stack (car token) (car action)))))

	(case (car action)

	  ((shift)
           (if (promise? value-stack)
               (set! value-stack (force value-stack)))
           (parser-loop (parser-context-cons g (cadr action) state-stack)
                        (parser-value-cons (cadr token) value-stack)
                        base-backtrack-stack
                        (lexer)
                        lexer
                        parse-error))

	  ((reduce)
           (let ((reduction-function (caadr action)) ; ugh
                 (remaining-reductions (cdadr action)))
             (reduction-function remaining-reductions
                                 parser-loop
                                 state-stack
                                 (if (promise? value-stack)
                                     value-stack
                                     (delay value-stack))
                                 backtrack-stack
                                 token
                                 lexer
                                 parse-error)))

	  ((error)
           (let ((backtrack-function (car backtrack-stack)))
             (backtrack-function parser-loop
                                 state-stack
                                 value-stack
                                 backtrack-stack
                                 token
                                 lexer
                                 parse-error))))))))

;;; lr1.scm ends here
