;;; yy.scm

;;	Copyright (C) 2002 Free Software Foundation, Inc.
;;
;; This file is part of GUILE
;;
;; GUILE is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.
;;
;; GUILE is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; Commentary:

;; This module abstracts module `(lang lr1)' in three ways:
;; - The grammar must not handle <eoi> tokens directly.
;; - The grammar cannot use symbol `%%YY' (anywhere).
;; - The grammar is expanded to handle YYACCEPT, YYABORT and YYERROR calls.

;;; Code:

(define-module (lang yy)
  :use-module (lang grammar)
  :use-module (lang lex)
  :use-module (lang lr1-debug)
  :export (yy-parse))

(for-each (lambda (msg)
            (display msg (current-error-port)))
          (list
           "WARNING: Module (lang yy) incomplete!\n"
           "(Feel free to hack on it and post patches, however.)\n"))


;;; Support

(define (splice-YY tree)
  (if (list? tree)
      (let ((head (car tree)))
        (cond ((and (symbol? head) (eq? 'YYERROR head))
               '(throw '%%YY 'error #t))
              ((and (symbol? head) (eq? 'YYABORT head))
               '(throw '%%YY 'abort #t))
              (else
               (append (if (and (symbol? head)
                                (eq? 'YYACCEPT head))
                           '(throw '%%YY 'accept)
                           (list (splice-YY head)))
                       (map splice-YY (cdr tree))))))
      tree))

(define (prep-grammar grammar)
  (map (lambda (p)
         (apply list
                (production-symbol p)
                (production-expansion p)
                (splice-YY (production-body p))))
       (append '()
               ;;`(($start$ (,(production-symbol (car grammar)) $eoi-shifter$)
               ;;           (YYACCEPT $1))
               ;;  ($eoi-shifter$ ($eoi-shifter$ <eoi>) #f)
               ;;  ($eoi-shifter$ () #f))
               grammar)))

(define (absorb-soft-eoi lexer)
  (letrec ((err (lambda (cont states values token lexer)
                  (cond
                   ((and (eof-token? token) (not (cadr token)))
                    (write-line "error handler continuing")
                    (cont states values (lexer) lexer err))
                   (else
                    (error (format #f "unexpected: ~S" token)))))))
    err))

(define (yy-parse grammar lexer module)
  (catch '%%YY
         (lambda ()
           (lr1-parse (prep-grammar grammar)
                      lexer
                      (absorb-soft-eoi lexer)
                      module))
         (lambda args
           (case (cadr args)
             ((accept) (if (= 3 (length args))
                           (caddr args)
                           (cddr args)))
             ((abort) #f)
             (else (error (format #f "unexpected %%YY: ~S" (cdr args))))))))

;;; yy.scm ends here
