;;; srfi-1.scm --- List Library

;;	Copyright (C) 2001, 2002 Free Software Foundation, Inc.
;;
;; This file is part of GUILE
;;
;; GUILE is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or
;; (at your option) any later version.
;;
;; GUILE is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this software; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
;; Boston, MA 02111-1307 USA
;;
;; As a special exception, the Free Software Foundation gives permission
;; for additional uses of the text contained in its release of GUILE.
;;
;; The exception is that, if you link the GUILE library with other files
;; to produce an executable, this does not by itself cause the
;; resulting executable to be covered by the GNU General Public License.
;; Your use of that executable is in no way restricted on account of
;; linking the GUILE library code into it.
;;
;; This exception does not however invalidate any other reasons why
;; the executable file might be covered by the GNU General Public License.
;;
;; This exception applies only to the code released by the
;; Free Software Foundation under the name GUILE.  If you copy
;; code from other Free Software Foundation releases into a copy of
;; GUILE, as the General Public License permits, the exception does
;; not apply to the code that you add in this way.  To avoid misleading
;; anyone as to the status of such modified files, you must delete
;; this exception notice from them.
;;
;; If you write modifications of your own for GUILE, it is your choice
;; whether to permit this exception to apply to your modifications.
;; If you do not wish that, delete this exception notice.

;;; Author: Martin Grabmueller <mgrabmue@cs.tu-berlin.de>
;;; Date: 2001-06-06

;;; Commentary:

;; This is an implementation of SRFI-1 (List Library).
;;
;; All procedures defined in SRFI-1, which are not already defined in
;; the Guile core library, are exported.  The procedures in this
;; implementation work, but they have not been tuned for speed or
;; memory usage.
;;
;; This module is fully documented in the Guile Reference Manual.

;;; Code:

(define-module (srfi srfi-1)
  :use-module (ice-9 receive))

(export
;;; Constructors
 ;; cons				<= in the core
 ;; list				<= in the core
 xcons
 ;; cons*				<= in the core
 ;; make-list				<= in the core
 list-tabulate
 ;; list-copy				<= in the core
 circular-list
 iota					; Extended.

;;; Predicates
 proper-list?
 circular-list?
 dotted-list?
 ;; pair?				<= in the core
 ;; null?				<= in the core
 null-list?
 not-pair?
 list=

;;; Selectors
 ;; car					<= in the core
 ;; cdr					<= in the core
 ;; caar				<= in the core
 ;; cadr				<= in the core
 ;; cdar				<= in the core
 ;; cddr				<= in the core
 ;; caaar				<= in the core
 ;; caadr				<= in the core
 ;; cadar				<= in the core
 ;; caddr				<= in the core
 ;; cdaar				<= in the core
 ;; cdadr				<= in the core
 ;; cddar				<= in the core
 ;; cdddr				<= in the core
 ;; caaaar				<= in the core
 ;; caaadr				<= in the core
 ;; caadar				<= in the core
 ;; caaddr				<= in the core
 ;; cadaar				<= in the core
 ;; cadadr				<= in the core
 ;; caddar				<= in the core
 ;; cadddr				<= in the core
 ;; cdaaar				<= in the core
 ;; cdaadr				<= in the core
 ;; cdadar				<= in the core
 ;; cdaddr				<= in the core
 ;; cddaar				<= in the core
 ;; cddadr				<= in the core
 ;; cdddar				<= in the core
 ;; cddddr				<= in the core
 ;; list-ref				<= in the core
 first
 second
 third
 fourth
 fifth
 sixth
 seventh
 eighth
 ninth
 tenth
 car+cdr
 take
 drop
 take-right
 drop-right
 take!
 drop-right!
 split-at
 split-at!
 last
 ;; last-pair				<= in the core

;;; Miscelleneous: length, append, concatenate, reverse, zip & count
 ;; length				<= in the core
 length+
 ;; append				<= in the core
 ;; append!				<= in the core
 concatenate
 concatenate!
 ;; reverse				<= in the core
 ;; reverse!				<= in the core
 append-reverse
 append-reverse!
 zip
 unzip1
 unzip2
 unzip3
 unzip4
 unzip5
 count

;;; Fold, unfold & map
 fold
 fold-right
 pair-fold
 pair-fold-right
 reduce
 reduce-right
 unfold
 unfold-right
 map					; Extended.
 for-each				; Extended.
 append-map
 append-map!
 map!
 map-in-order				; Extended.
 pair-for-each
 filter-map

;;; Filtering & partitioning
 filter
 partition
 remove
 filter!
 partition!
 remove!

;;; Searching
 find
 find-tail
 take-while
 take-while!
 drop-while
 span
 span!
 break
 break!
 any
 every
 list-index				; Extended.
 member					; Extended.
 ;; memq				<= in the core
 ;; memv				<= in the core

;;; Deletion
 delete					; Extended.
 delete!				; Extended.
 delete-duplicates
 delete-duplicates!

;;; Association lists
 assoc					; Extended.
 ;; assq				<= in the core
 ;; assv				<= in the core
 alist-cons
 alist-copy
 alist-delete
 alist-delete!

;;; Set operations on lists
 lset<=
 lset=
 lset-adjoin
 lset-union
 lset-intersection
 lset-difference
 lset-xor
 lset-diff+intersection
 lset-union!
 lset-intersection!
 lset-difference!
 lset-xor!
 lset-diff+intersection!

;;; Primitive side-effects
 ;; set-car!				<= in the core
 ;; set-cdr!				<= in the core
 )

(cond-expand-provide (current-module) '(srfi-1))

;;; Constructors

;; Like @code{cons}, but with interchanged arguments.  Useful mostly when
;; passed to higher-order procedures.
(define (xcons d a)
  (cons a d))

;; internal helper, similar to (scsh utilities) check-arg.
(define (check-arg-type pred arg caller)
  (if (pred arg)
      arg
      (scm-error 'wrong-type-arg caller
		 "Wrong type argument: ~S" (list arg) '())))

;; the srfi spec doesn't seem to forbid inexact integers.
(define (non-negative-integer? x) (and (integer? x) (>= x 0)))

;; Return an @var{n}-element list, where each list element is produced by
;; applying the procedure @var{init-proc} to the corresponding list
;; index.  The order in which @var{init-proc} is applied to the indices
;; is not specified.
(define (list-tabulate n init-proc)
  (check-arg-type non-negative-integer? n "list-tabulate")
  (let lp ((n n) (acc '()))
    (if (<= n 0)
      acc
      (lp (- n 1) (cons (init-proc (- n 1)) acc)))))

;; Return a circular list containing the given arguments @var{elt1}
;; @var{elt2} @dots{}.
(define (circular-list elt1 . rest)
  (let ((start (cons elt1 '())))
    (let lp ((r rest) (p start))
      (if (null? r)
	(begin
	  (set-cdr! p start)
	  start)
	(begin
	  (set-cdr! p (cons (car r) '()))
	  (lp (cdr r) (cdr p)))))))

(define (iota count . rest)
  (check-arg-type non-negative-integer? count "iota")
  (let ((start (if (pair? rest) (car rest) 0))
	(step (if (and (pair? rest) (pair? (cdr rest))) (cadr rest) 1)))
    (let lp ((n 0) (acc '()))
      (if (= n count)
	(reverse! acc)
	(lp (+ n 1) (cons (+ start (* n step)) acc))))))

;;; Predicates

;; Return @code{#t} if @var{obj} is a proper list, that is a finite list,
;; terminated with the empty list.  Otherwise, return @code{#f}.
(define (proper-list? x)
  (list? x))

;; Return @code{#t} if @var{obj} is a circular list, otherwise return
;; @code{#f}.
(define (circular-list? x)
  (if (not-pair? x)
    #f
    (let lp ((hare (cdr x)) (tortoise x))
      (if (not-pair? hare)
	#f
	(let ((hare (cdr hare)))
	  (if (not-pair? hare)
	    #f
	    (if (eq? hare tortoise)
	      #t
	      (lp (cdr hare) (cdr tortoise)))))))))

;; Return @code{#t} if @var{obj} is a dotted list, return @code{#f}
;; otherwise.  A dotted list is a finite list which is not terminated by
;; the empty list, but some other value.
(define (dotted-list? x)
  (cond
    ((null? x) #f)
    ((not-pair? x) #t)
    (else
     (let lp ((hare (cdr x)) (tortoise x))
       (cond
	 ((null? hare) #f)
	 ((not-pair? hare) #t)
	 (else
	  (let ((hare (cdr hare)))
	    (cond
	      ((null? hare) #f)
	      ((not-pair? hare) #t)
	      ((eq? hare tortoise) #f)
	      (else
	       (lp (cdr hare) (cdr tortoise)))))))))))

;; Return @code{#t} if @var{lst} is the empty list @code{()}, @code{#f}
;; otherwise.  If something else than a proper or circular list is passed
;; as @var{lst}, an error is signalled.  This procedure is recommended
;; for checking for the end of a list in contexts where dotted lists are
;; not allowed.
(define (null-list? x)
  (cond
    ((proper-list? x)
     (null? x))
    ((circular-list? x)
     #f)
    (else
     (error "not a proper list in null-list?"))))

;; Return @code{#t} is @var{obj} is not a pair, @code{#f} otherwise.
;; This is shorthand notation @code{(not (pair? @var{obj}))} and is
;; supposed to be used for end-of-list checking in contexts where dotted
;; lists are allowed.
(define (not-pair? x)
  (not (pair? x)))

;; Return @code{#t} if all argument lists are equal, @code{#f} otherwise.
;; List equality is determined by testing whether all lists have the same
;; length and the corresponding elements are equal in the sense of the
;; equality predicate @var{elt=}.  If no or only one list is given,
;; @code{#t} is returned.
(define (list= elt= . rest)
  (define (lists-equal a b)
    (let lp ((a a) (b b))
      (cond ((null? a)
	     (null? b))
	    ((null? b)
	     #f)
	    (else
	     (and (elt= (car a) (car b))
		  (lp (cdr a) (cdr b)))))))
  (or (null? rest)
      (let ((first (car rest)))
	(let lp ((lists rest))
	  (or (null? lists)
	      (and (lists-equal first (car lists))
		   (lp (cdr lists))))))))

;;; Selectors

;; These are synonyms for @code{car}, @code{cadr}, @code{caddr}, @dots{}.
(define (first pair) (car pair))        ; (syntax friendly to doc-snarf)
(define second cadr)
(define third caddr)
(define fourth cadddr)
(define (fifth x) (car (cddddr x)))
(define (sixth x) (cadr (cddddr x)))
(define (seventh x) (caddr (cddddr x)))
(define (eighth x) (cadddr (cddddr x)))
(define (ninth x) (car (cddddr (cddddr x))))
(define (tenth x) (cadr (cddddr (cddddr x))))

;; Return two values, the @sc{car} and the @sc{cdr} of @var{pair}.
(define (car+cdr pair) (values (car pair) (cdr pair)))

;; Return a list containing the first @var{i} elements of @var{lst}.
;;
;; @code{take!} may modify the structure of the argument list @var{lst}
;; in order to produce the result.
(define (take lst i)
  (let lp ((n i) (l lst) (acc '()))
    (if (<= n 0)
      (reverse! acc)
      (lp (- n 1) (cdr l) (cons (car l) acc)))))

;; Return a list containing all but the first @var{i} elements of
;; @var{lst}.
(define (drop lst i)
  (let lp ((n i) (l lst))
    (if (<= n 0)
      l
      (lp (- n 1) (cdr l)))))

;; Return the a list containing the @var{i} last elements of @var{flist}.
(define (take-right flist i)
  (let lp ((n i) (l flist))
    (if (<= n 0)
      (let lp0 ((s flist) (l l))
	(if (null? l)
	  s
	  (lp0 (cdr s) (cdr l))))
      (lp (- n 1) (cdr l)))))

;; Return the a list containing all but the @var{i} last elements of
;; @var{flist}.
;;
;; @code{drop-right!} may modify the structure of the argument list
;; @var{flist} in order to produce the result.
(define (drop-right flist i)
  (let lp ((n i) (l flist))
    (if (<= n 0)
      (let lp0 ((s flist) (l l) (acc '()))
	(if (null? l)
	  (reverse! acc)
	  (lp0 (cdr s) (cdr l) (cons (car s) acc))))
      (lp (- n 1) (cdr l)))))

(define (take! x i)
  (if (<= i 0)
    '()
    (let lp ((n (- i 1)) (l x))
      (if (<= n 0)
	(begin
	  (set-cdr! l '())
	  x)
	(lp (- n 1) (cdr l))))))

(define (drop-right! flist i)
  (if (<= i 0)
    flist
    (let lp ((n (+ i 1)) (l flist))
      (if (<= n 0)
	(let lp0 ((s flist) (l l))
	  (if (null? l)
	    (begin
	      (set-cdr! s '())
	      flist)
	    (lp0 (cdr s) (cdr l))))
	(if (null? l)
	  '()
	  (lp (- n 1) (cdr l)))))))

;; Return two values, a list containing the first @var{i} elements of the
;; list @var{lst} and a list containing the remaining elements.
;;
;; @code{split-at!} may modify the structure of the argument list
;; @var{lst} in order to produce the result.
(define (split-at lst i)
  (let lp ((l lst) (n i) (acc '()))
    (if (<= n 0)
      (values (reverse! acc) l)
      (lp (cdr l) (- n 1) (cons (car l) acc)))))

(define (split-at! x i)
  (if (<= i 0)
    (values '() x)
    (let lp ((l x) (n (- i 1)))
      (if (<= n 0)
	(let ((tmp (cdr l)))
	  (set-cdr! l '())
	  (values x tmp))
	(lp (cdr l) (- n 1))))))

;; Return the last element of the non-empty, finite list @var{lst}.
(define (last lst)
  (car (last-pair lst)))

;;; Miscelleneous: length, append, concatenate, reverse, zip & count

;; Return the length of the argument list @var{lst}.  When @var{lst} is a
;; circular list, @code{#f} is returned.
(define (length+ clist)
  (if (null? clist)
    0
    (let lp ((hare (cdr clist)) (tortoise clist) (l 1))
      (if (null? hare)
	l
	(let ((hare (cdr hare)))
	  (if (null? hare)
	    (+ l 1)
	    (if (eq? hare tortoise)
	      #f
	      (lp (cdr hare) (cdr tortoise) (+ l 2)))))))))

;; Construct a list by appending all lists in @var{list-of-lists}.
;;
;; @code{concatenate!} may modify the structure of the given lists in
;; order to produce the result.
(define (concatenate l-o-l)
  (let lp ((l l-o-l) (acc '()))
    (if (null? l)
      (reverse! acc)
      (let lp0 ((ll (car l)) (acc acc))
	(if (null? ll)
	  (lp (cdr l) acc)
	  (lp0 (cdr ll) (cons (car ll) acc)))))))

(define (concatenate! l-o-l)
  (let lp0 ((l-o-l l-o-l))
    (cond
      ((null? l-o-l)
       '())
      ((null? (car l-o-l))
       (lp0 (cdr l-o-l)))
      (else
       (let ((result (car l-o-l)) (tail (last-pair (car l-o-l))))
	 (let lp ((l (cdr l-o-l)) (ntail tail))
	   (if (null? l)
	     result
	     (begin
	       (set-cdr! ntail (car l))
	       (lp (cdr l) (last-pair ntail))))))))))

;; Reverse @var{rev-head}, append @var{tail} and return the result.  This
;; is equivalent to @code{(append (reverse @var{rev-head}) @var{tail})},
;; but more efficient.
;;
;; @code{append-reverse!} may modify @var{rev-head} in order to produce
;; the result.
(define (append-reverse rev-head tail)
  (let lp ((l rev-head) (acc tail))
    (if (null? l)
      acc
      (lp (cdr l) (cons (car l) acc)))))

(define (append-reverse! rev-head tail)
  (append-reverse rev-head tail))	; XXX:optimize

;; Return a list as long as the shortest of the argument lists, where
;; each element is a list.  The first list contains the first elements of
;; the argument lists, the second list contains the second elements, and
;; so on.
(define (zip clist1 . rest)
  (let lp ((l (cons clist1 rest)) (acc '()))
    (if (any null? l)
      (reverse! acc)
      (lp (map1 cdr l) (cons (map1 car l) acc)))))


;; @code{unzip1} takes a list of lists, and returns a list containing the
;; first elements of each list, @code{unzip2} returns two lists, the
;; first containing the first elements of each lists and the second
;; containing the second elements of each lists, and so on.
(define (unzip1 lst)
  (map1 first lst))
(define (unzip2 l)
  (values (map1 first l) (map1 second l)))
(define (unzip3 l)
  (values (map1 first l) (map1 second l) (map1 third l)))
(define (unzip4 l)
  (values (map1 first l) (map1 second l) (map1 third l) (map1 fourth l)))
(define (unzip5 l)
  (values (map1 first l) (map1 second l) (map1 third l) (map1 fourth l)
	  (map1 fifth l)))

(define (count pred clist1 . rest)
  (if (null? rest)
      (count1 pred clist1)
      (let lp ((lists (cons clist1 rest)))
	(cond ((any1 null? lists)
	       0)
	      (else
	       (if (apply pred (map1 car lists))
		 (+ 1 (lp (map1 cdr lists)))
		 (lp (map1 cdr lists))))))))

(define (count1 pred clist)
  (let lp ((result 0) (rest clist))
    (if (null? rest)
	result
	(if (pred (car rest))
	    (lp (+ 1 result) (cdr rest))
	    (lp result (cdr rest))))))

;;; Fold, unfold & map

;; Fold the procedure @var{kons} across all elements of @var{lst1},
;; @var{lst2}, @dots{}.  Produce the result of
;;
;; @code{(@var{kons} @var{en1} @var{en2} @dots{} (@var{kons} @var{e21}
;; @var{e22} (@var{kons} @var{e11} @var{e12} @var{knil})))},
;;
;; if @var{enm} are the elements of the lists @var{lst1}, @var{lst2},
;; @dots{}.
(define (fold kons knil list1 . rest)
  (if (null? rest)
      (let f ((knil knil) (list1 list1))
	(if (null? list1)
	    knil
	    (f (kons (car list1) knil) (cdr list1))))
      (let f ((knil knil) (lists (cons list1 rest)))
	(if (any null? lists)
	    knil
	    (let ((cars (map1 car lists))
		  (cdrs (map1 cdr lists)))
	      (f (apply kons (append! cars (list knil))) cdrs))))))

;; Similar to @code{fold}, but applies @var{kons} in right-to-left order
;; to the list elements, that is:
;;
;; @code{(@var{kons} @var{e11} @var{e12}(@var{kons} @var{e21}
;; @var{e22}  @dots{} (@var{kons} @var{en1} @var{en2} @var{knil})))},
(define (fold-right kons knil clist1 . rest)
  (if (null? rest)
    (let f ((list1 clist1))
      (if (null? list1)
	knil
	(kons (car list1) (f (cdr list1)))))
    (let f ((lists (cons clist1 rest)))
      (if (any null? lists)
	knil
	(apply kons (append! (map1 car lists) (list (f (map1 cdr lists)))))))))

;; Like @code{fold}, but apply @var{kons} to the pairs of the list
;; instead of the list elements.
(define (pair-fold kons knil clist1 . rest)
  (if (null? rest)
      (let f ((knil knil) (list1 clist1))
	(if (null? list1)
	    knil
	    (let ((tail (cdr list1)))
	    (f (kons list1 knil) tail))))
      (let f ((knil knil) (lists (cons clist1 rest)))
	(if (any null? lists)
	    knil
	    (let ((tails (map1 cdr lists)))
	      (f (apply kons (append! lists (list knil))) tails))))))

;; Like @code{fold-right}, but apply @var{kons} to the pairs of the list
;; instead of the list elements.
(define (pair-fold-right kons knil clist1 . rest)
  (if (null? rest)
    (let f ((list1 clist1))
      (if (null? list1)
	knil
	(kons list1 (f (cdr list1)))))
    (let f ((lists (cons clist1 rest)))
      (if (any null? lists)
	knil
	(apply kons (append! lists (list (f (map1 cdr lists)))))))))

;; @code{unfold} is defined as follows:
;;
;; @lisp
;; (unfold p f g seed) =
;;    (if (p seed) (tail-gen seed)
;;        (cons (f seed)
;;              (unfold p f g (g seed))))
;; @end lisp
;;
;; @table @var
;; @item p
;; Determines when to stop unfolding.
;;
;; @item f
;; Maps each seed value to the corresponding list element.
;;
;; @item g
;; Maps each seed value to next seed valu.
;;
;; @item seed
;; The state value for the unfold.
;;
;; @item tail-gen
;; Creates the tail of the list; defaults to @code{(lambda (x) '())}.
;; @end table
;;
;; @var{g} produces a series of seed values, which are mapped to list
;; elements by @var{f}.  These elements are put into a list in
;; left-to-right order, and @var{p} tells when to stop unfolding.
(define (unfold p f g seed . rest)
  (let ((tail-gen (if (pair? rest)
		      (if (pair? (cdr rest))
			  (scm-error 'wrong-number-of-args
				     "unfold" "too many arguments" '() '())
			  (car rest))
		      (lambda (x) '()))))
    (let uf ((seed seed))
      (if (p seed)
	  (tail-gen seed)
	  (cons (f seed)
		(uf (g seed)))))))

;; Construct a list with the following loop.
;;
;; @lisp
;; (let lp ((seed seed) (lis tail))
;;    (if (p seed) lis
;;        (lp (g seed)
;;            (cons (f seed) lis))))
;; @end lisp
;;
;; @table @var
;; @item p
;; Determines when to stop unfolding.
;;
;; @item f
;; Maps each seed value to the corresponding list element.
;;
;; @item g
;; Maps each seed value to next seed valu.
;;
;; @item seed
;; The state value for the unfold.
;;
;; @item tail-gen
;; Creates the tail of the list; defaults to @code{(lambda (x) '())}.
;; @end table
(define (unfold-right p f g seed . rest)
  (let ((tail (if (pair? rest)
		  (if (pair? (cdr rest))
		      (scm-error 'wrong-number-of-args
				     "unfold-right" "too many arguments" '()
				     '())
		      (car rest))
		      '())))
    (let uf ((seed seed) (lis tail))
      (if (p seed)
	  lis
	  (uf (g seed) (cons (f seed) lis))))))

;; @code{reduce} is a variant of @code{fold}.  If @var{lst} is
;; @code{()}, @var{ridentity} is returned.  Otherwise, @code{(fold (car
;; @var{lst}) (cdr @var{lst}))} is returned.
(define (reduce f ridentity lst)
  (fold f ridentity lst))

;; This is the @code{fold-right} variant of @var{reduce}.
(define (reduce-right f ridentity lst)
  (fold-right f ridentity lst))


;; Internal helper procedure.  Map `f' over the single list `ls'.

(define (map1 f ls)
  (if (null? ls)
      ls
      (let ((ret (list (f (car ls)))))
        (let lp ((ls (cdr ls)) (p ret))         ; tail pointer
          (if (null? ls)
              ret
              (begin
                (set-cdr! p (list (f (car ls))))
                (lp (cdr ls) (cdr p))))))))

;; Map the procedure over the list(s) @var{lst1}, @var{lst2}, @dots{} and
;; return a list containing the results of the procedure applications.
;; This procedure is extended with respect to R5RS, because the argument
;; lists may have different lengths.  The result list will have the same
;; length as the shortest argument lists.  The order in which @var{f}
;; will be applied to the list element(s) is not specified.
(define (map f list1 . rest)
  (if (null? rest)
    (map1 f list1)
    (let lp ((l (cons list1 rest)))
      (if (any1 null? l)
	'()
	(cons (apply f (map1 car l)) (lp (map1 cdr l)))))))

(define map-in-order map)

;; Apply the procedure @var{f} to each pair of corresponding elements of
;; the list(s) @var{lst1}, @var{lst2}, @dots{}.  The return value is not
;; specified.  This procedure is extended with respect to R5RS, because
;; the argument lists may have different lengths.  The shortest argument
;; list determines the number of times @var{f} is called.  @var{f} will
;; be applied to the list elements in left-to-right order.
(define (for-each f list1 . rest)
  (if (null? rest)
    (let lp ((l list1))
      (if (null? l)
	(if #f #f)			; Return unspecified value.
	(begin
	  (f (car l))
	  (lp (cdr l)))))
    (let lp ((l (cons list1 rest)))
      (if (any1 null? l)
	(if #f #f)
	(begin
	  (apply f (map1 car l))
	  (lp (map1 cdr l)))))))

;; Equivalent to
;;
;; @lisp
;; (apply append (map f clist1 clist2 ...))
;; @end lisp
;;
;; and
;;
;; @lisp
;; (apply append! (map f clist1 clist2 ...))
;; @end lisp
;;
;; Map @var{f} over the elements of the lists, just as in the @code{map}
;; function. However, the results of the applications are appended
;; together to make the final result. @code{append-map} uses
;; @code{append} to append the results together; @code{append-map!} uses
;; @code{append!}.
;;
;; The dynamic order in which the various applications of @var{f} are
;; made is not specified.
(define (append-map f clist1 . rest)
  (if (null? rest)
    (let lp ((l clist1))
      (if (null? l)
	'()
	(append (f (car l)) (lp (cdr l)))))
    (let lp ((l (cons clist1 rest)))
      (if (any1 null? l)
	'()
	(append (apply f (map1 car l)) (lp (map1 cdr l)))))))


(define (append-map! f clist1 . rest)
  (if (null? rest)
    (let lp ((l clist1))
      (if (null? l)
	'()
	(append! (f (car l)) (lp (cdr l)))))
    (let lp ((l (cons clist1 rest)))
      (if (any1 null? l)
	'()
	(append! (apply f (map1 car l)) (lp (map1 cdr l)))))))

;; Linear-update variant of @code{map} -- @code{map!} is allowed, but not
;; required, to alter the cons cells of @var{lst1} to construct the
;; result list.
;;
;; The dynamic order in which the various applications of @var{f} are
;; made is not specified. In the n-ary case, @var{lst2}, @var{lst3},
;; @dots{} must have at least as many elements as @var{lst1}.
(define (map! f list1 . rest)
  (if (null? rest)
    (let lp ((l list1))
      (if (null? l)
	'()
	(begin
	  (set-car! l (f (car l)))
	  (set-cdr! l (lp (cdr l)))
	  l)))
    (let lp ((l (cons list1 rest)) (res list1))
      (if (any1 null? l)
	'()
	(begin
	  (set-car! res (apply f (map1 car l)))
	  (set-cdr! res (lp (map1 cdr l) (cdr res)))
	  res)))))

;; Like @code{for-each}, but applies the procedure @var{f} to the pairs
;; from which the argument lists are constructed, instead of the list
;; elements.  The return value is not specified.
(define (pair-for-each f clist1 . rest)
  (if (null? rest)
    (let lp ((l clist1))
      (if (null? l)
	(if #f #f)
	(begin
	  (f l)
	  (lp (cdr l)))))
    (let lp ((l (cons clist1 rest)))
      (if (any1 null? l)
	(if #f #f)
	(begin
	  (apply f l)
	  (lp (map1 cdr l)))))))

;; Like @code{map}, but only results from the applications of @var{f}
;; which are true are saved in the result list.
(define (filter-map f clist1 . rest)
  (if (null? rest)
    (let lp ((l clist1))
      (if (null? l)
	'()
	(let ((res (f (car l))))
	  (if res
	    (cons res (lp (cdr l)))
	    (lp (cdr l))))))
    (let lp ((l (cons clist1 rest)))
      (if (any1 null? l)
	'()
	(let ((res (apply f (map1 car l))))
	  (if res
	    (cons res (lp (map1 cdr l)))
	    (lp (map1 cdr l))))))))

;;; Filtering & partitioning

;; Return a list containing all elements from @var{lst} which satisfy the
;; predicate @var{pred}.  The elements in the result list have the same
;; order as in @var{lst}.  The order in which @var{pred} is applied to
;; the list elements is not specified.
;;
;; @code{filter!} is allowed, but not required to modify the structure of
;; the list.
(define (filter pred list)
  (check-arg-type list? list "filter")  ; reject circular lists.
  (letrec ((filiter (lambda (pred rest result)
		      (if (null? rest)
			  (reverse! result)
			  (filiter pred (cdr rest)
				   (cond ((pred (car rest))
					  (cons (car rest) result))
					 (else
					  result)))))))
    (filiter pred list '())))

;; Return two lists, one containing all elements from @var{lst} which
;; satisfy the predicate @var{pred}, and one list containing the elements
;; which do not satisfy the predicated.  The elements in the result lists
;; have the same order as in @var{lst}.  The order in which @var{pred} is
;; applied to the list elements is not specified.
;;
;; @code{partition!} is allowed, but not required to modify the structure of
;; the input list.
(define (partition pred list)
  (if (null? list)
    (values '() '())
    (if (pred (car list))
      (receive (in out) (partition pred (cdr list))
	       (values (cons (car list) in) out))
      (receive (in out) (partition pred (cdr list))
	       (values in (cons (car list) out))))))

;; Return a list containing all elements from @var{lst} which do not
;; satisfy the predicate @var{pred}.  The elements in the result list
;; have the same order as in @var{lst}.  The order in which @var{pred} is
;; applied to the list elements is not specified.
;;
;; @code{remove!} is allowed, but not required to modify the structure of
;; the input list.
(define (remove pred list)
  (filter (lambda (x) (not (pred x))) list))

(define (filter! pred list)
  (filter pred list))			; XXX:optimize

(define (partition! pred list)
  (partition pred list))		; XXX:optimize

(define (remove! pred list)
  (remove pred list))			; XXX:optimize

;;; Searching

;; Return the first element of @var{lst} which satisfies the predicate
;; @var{pred} and @code{#f} if no such element is found.
(define (find pred clist)
  (if (null? clist)
    #f
    (if (pred (car clist))
      (car clist)
      (find pred (cdr clist)))))

;; Return the first pair of @var{lst} whose @sc{car} satisfies the
;; predicate @var{pred} and @code{#f} if no such element is found.
(define (find-tail pred clist)
  (if (null? clist)
    #f
    (if (pred (car clist))
      clist
      (find-tail pred (cdr clist)))))

;; Return the longest initial prefix of @var{lst} whose elements all
;; satisfy the predicate @var{pred}.
;;
;; @code{take-while!} is allowed, but not required to modify the input
;; list while producing the result.
(define (take-while pred ls)
  (cond ((null? ls) '())
        ((not (pred (car ls))) '())
        (else
         (let ((result (list (car ls))))
           (let lp ((ls (cdr ls)) (p result))
             (cond ((null? ls) result)
                   ((not (pred (car ls))) result)
                   (else
                    (set-cdr! p (list (car ls)))
                    (lp (cdr ls) (cdr p)))))))))

(define (take-while! pred clist)
  (take-while pred clist))		; XXX:optimize

;; Drop the longest initial prefix of @var{lst} whose elements all
;; satisfy the predicate @var{pred}.
(define (drop-while pred clist)
  (if (null? clist)
    '()
    (if (pred (car clist))
      (drop-while pred (cdr clist))
      clist)))

;; @code{span} splits the list @var{lst} into the longest initial prefix
;; whose elements all satisfy the predicate @var{pred}, and the remaining
;; tail.  @code{break} inverts the sense of the predicate.
;;
;; @code{span!} and @code{break!} are allowed, but not required to modify
;; the structure of the input list @var{lst} in order to produce the
;; result.
(define (span pred clist)
  (if (null? clist)
    (values '() '())
    (if (pred (car clist))
      (receive (first last) (span pred (cdr clist))
	       (values (cons (car clist) first) last))
      (values '() clist))))

(define (span! pred list)
  (span pred list))			; XXX:optimize

(define (break pred clist)
  (if (null? clist)
    (values '() '())
    (if (pred (car clist))
      (values '() clist)
      (receive (first last) (break pred (cdr clist))
	       (values (cons (car clist) first) last)))))

(define (break! pred list)
  (break pred list))			; XXX:optimize

;; Apply @var{pred} across the lists and return a true value if the
;; predicate returns true for any of the list elements(s); return
;; @code{#f} otherwise.  The true value returned is always the result of
;; the first successful application of @var{pred}.
(define (any pred ls . lists)
  (if (null? lists)
      (any1 pred ls)
      (let lp ((lists (cons ls lists)))
	(cond ((any1 null? lists)
	       #f)
	      ((any1 null? (map1 cdr lists))
	       (apply pred (map1 car lists)))
	      (else
	       (or (apply pred (map1 car lists)) (lp (map1 cdr lists))))))))

(define (any1 pred ls)
  (let lp ((ls ls))
    (cond ((null? ls)
	   #f)
	  ((null? (cdr ls))
	   (pred (car ls)))
	  (else
	   (or (pred (car ls)) (lp (cdr ls)))))))

;; Apply @var{pred} across the lists and return a true value if the
;; predicate returns true for every of the list elements(s); return
;; @code{#f} otherwise.  The true value returned is always the result of
;; the final successful application of @var{pred}.
(define (every pred ls . lists)
  (if (null? lists)
      (every1 pred ls)
      (let lp ((lists (cons ls lists)))
	(cond ((any1 null? lists)
	       #t)
	      ((any1 null? (map1 cdr lists))
	       (apply pred (map1 car lists)))
	      (else
	       (and (apply pred (map1 car lists)) (lp (map1 cdr lists))))))))

(define (every1 pred ls)
  (let lp ((ls ls))
    (cond ((null? ls)
	   #t)
	  ((null? (cdr ls))
	   (pred (car ls)))
	  (else
	   (and (pred (car ls)) (lp (cdr ls)))))))

;; Return the index of the leftmost element that satisfies @var{pred}.
(define (list-index pred clist1 . rest)
  (if (null? rest)
    (let lp ((l clist1) (i 0))
      (if (null? l)
	#f
	(if (pred (car l))
	  i
	  (lp (cdr l) (+ i 1)))))
    (let lp ((lists (cons clist1 rest)) (i 0))
      (cond ((any1 null? lists)
	     #f)
	    ((apply pred (map1 car lists)) i)
	    (else
	     (lp (map1 cdr lists) (+ i 1)))))))

;; Return the first sublist of @var{lst} whose @sc{car} is equal to
;; @var{x}.  If @var{x} does no appear in @var{lst}, return @code{#f}.
;; Equality is determined by the equality predicate @var{=}, or
;; @code{equal?} if @var{=} is not given.
(define (member x list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (let lp ((l list))
      (if (null? l)
	#f
	(if (l= x (car l))
	  l
	  (lp (cdr l)))))))

;;; Deletion

;; Return a list containing all elements from @var{lst}, but without the
;; elements equal to @var{x}.  Equality is determined by the equality
;; predicate @var{=}, which defaults to @code{equal?} if not given.
;;
;; @code{delete!} is allowed, but not required to modify the structure of
;; the argument list in order to produce the result.
(define (delete x list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (let lp ((l list))
      (if (null? l)
	'()
	(if (l= (car l) x)
	  (lp (cdr l))
	  (cons (car l) (lp (cdr l))))))))

(define (delete! x list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (delete x list l=)))		; XXX:optimize

;; Return a list containing all elements from @var{lst}, but without
;; duplicate elements.  Equality of elements is determined by the
;; equality predicate @var{=}, which defaults to @code{equal?} if not
;; given.
;;
;; @code{delete-duplicates!} is allowed, but not required to modify the
;; structure of the argument list in order to produce the result.
(define (delete-duplicates list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (let lp0 ((l1 list))
      (if (null? l1)
	'()
	(if (let lp1 ((l2 (cdr l1)))
	      (if (null? l2)
		#f
		(if (l= (car l1) (car l2))
		  #t
		  (lp1 (cdr l2)))))
	  (lp0 (cdr l1))
	  (cons (car l1) (lp0 (cdr l1))))))))

(define (delete-duplicates list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (let lp ((list list))
      (if (null? list)
	'()
	(cons (car list) (lp (delete (car list) (cdr list) l=)))))))

(define (delete-duplicates! list . rest)
  (let ((l= (if (pair? rest) (car rest) equal?)))
    (delete-duplicates list l=)))	; XXX:optimize

;;; Association lists

(define (assoc key alist . rest)
  (let ((k= (if (pair? rest) (car rest) equal?)))
    (let lp ((a alist))
      (if (null? a)
	#f
	(if (k= key (caar a))
	  (car a)
	  (lp (cdr a)))))))

;; Equivalent to
;;
;; @lisp
;; (cons (cons @var{key} @var{datum}) @var{alist})
;; @end lisp
;;
;; This procedure is used to coons a new pair onto an existing
;; association list.
(define (alist-cons key datum alist)
  (acons key datum alist))

;; Return a newly allocated copy of @var{alist}, that means that the
;; spine of the list as well as the pairs are copied.
(define (alist-copy alist)
  (let lp ((a alist))
    (if (null? a)
      '()
      (acons (caar a) (cdar a) (lp (cdr a))))))

;; Return a list containing the pairs of @var{alist}, but without the
;; pairs whose @sc{cars} are equal to @var{key}.  Equality is determined
;; by @var{=}, which defaults to @code{equal?} if not given.
;;
;; @code{alist-delete!} is allowed, but not required to modify the
;; structure of the list @var{alist} in order to produce the result.
(define (alist-delete key alist . rest)
  (let ((k= (if (pair? rest) (car rest) equal?)))
    (let lp ((a alist))
      (if (null? a)
	'()
	(if (k= (caar a) key)
	  (lp (cdr a))
	  (cons (car a) (lp (cdr a))))))))

(define (alist-delete! key alist . rest)
  (let ((k= (if (pair? rest) (car rest) equal?)))
    (alist-delete key alist k=)))	; XXX:optimize

;;; Set operations on lists

;; Return @code{#t} if every @var{listi} is a subset of @var{listi+1},
;; otherwise return @code{#f}.  Returns @code{#t} if called with less
;; than two arguments. @var{=} is used for testing element equality.
(define (lset<= = . rest)
  (if (null? rest)
    #t
    (let lp ((f (car rest)) (r (cdr rest)))
      (or (null? r)
	  (and (every (lambda (el) (member el (car r) =)) f)
	       (lp (car r) (cdr r)))))))

;; Return @code{#t} if all argument lists are equal. @var{=} is used for
;; testing element equality.
(define (lset= = list1 . rest)
  (if (null? rest)
    #t
    (let lp ((f list1) (r rest))
      (or (null? r)
	  (and (every (lambda (el) (member el (car r) =)) f)
	       (every (lambda (el) (member el f =)) (car r))
	       (lp (car r) (cdr r)))))))

;; Add all @var{elts} to the list @var{list}, suppressing duplicates and
;; return the resulting list.  @code{lset-adjoin!} is allowed, but not
;; required to modify its first argument. @var{=} is used for testing
;; element equality.
(define (lset-adjoin = list . rest)
  (let lp ((l rest) (acc list))
    (if (null? l)
      acc
      (if (member (car l) acc)
	(lp (cdr l) acc)
	(lp (cdr l) (cons (car l) acc))))))

;; Return the union of all argument list sets.  The union is the set of
;; all elements which appear in any of the argument sets.
;; @code{lset-union!} is allowed, but not required to modify its first
;; argument. @var{=} is used for testing element equality.
(define (lset-union = . rest)
  (let lp0 ((l rest) (acc '()))
    (if (null? l)
      (reverse! acc)
      (let lp1 ((ll (car l)) (acc acc))
	(if (null? ll)
	  (lp0 (cdr l) acc)
	  (if (member (car ll) acc =)
	    (lp1 (cdr ll) acc)
	    (lp1 (cdr ll) (cons (car ll) acc))))))))

;; Return the intersection of all argument list sets.  The intersection
;; is the set containing all elements which appear in all argument sets.
;; @code{lset-intersection!} is allowed, but not required to modify its
;; first argument. @var{=} is used for testing element equality.
(define (lset-intersection = list1 . rest)
  (let lp ((l list1) (acc '()))
    (if (null? l)
      (reverse! acc)
      (if (every (lambda (ll) (member (car l) ll =)) rest)
	(lp (cdr l) (cons (car l) acc))
	(lp (cdr l) acc)))))

;; Return the difference of all argument list sets.  The difference is
;; the the set containing all elements of the first list which do not
;; appear in the other lists.  @code{lset-difference!}  is allowed, but
;; not required to modify its first argument. @var{=} is used for testing
;; element equality.
(define (lset-difference = list1 . rest)
  (if (null? rest)
    list1
    (let lp ((l list1) (acc '()))
      (if (null? l)
	(reverse! acc)
	(if (any (lambda (ll) (member (car l) ll =)) rest)
	  (lp (cdr l) acc)
	  (lp (cdr l) (cons (car l) acc)))))))

;(define (fold kons knil list1 . rest)

;; Return the set containing all elements which appear in the first
;; argument list set, but not in the second; or, more generally: which
;; appear in an odd number of sets.  @code{lset-xor!}  is allowed, but
;; not required to modify its first argument. @var{=} is used for testing
;; element equality.
(define (lset-xor = . rest)
  (fold (lambda (lst res)
	  (let lp ((l lst) (acc '()))
	    (if (null? l)
	      (let lp0 ((r res) (acc acc))
		(if (null? r)
		  (reverse! acc)
		  (if (member (car r) lst =)
		    (lp0 (cdr r) acc)
		    (lp0 (cdr r) (cons (car r) acc)))))
	      (if (member (car l) res =)
		(lp (cdr l) acc)
		(lp (cdr l) (cons (car l) acc))))))
	'()
	rest))

;; Return two values, the difference and the intersection of the argument
;; list sets. This works like a combination of @code{lset-difference} and
;; @code{lset-intersection}, but is more efficient.
;; @code{lset-diff+intersection!}  is allowed, but not required to modify
;; its first argument. @var{=} is used for testing element equality.  You
;; have to use some means to deal with the multiple values these
;; procedures return (@pxref{Multiple Values}).
(define (lset-diff+intersection = list1 . rest)
  (let lp ((l list1) (accd '()) (acci '()))
    (if (null? l)
      (values (reverse! accd) (reverse! acci))
      (let ((appears (every (lambda (ll) (member (car l) ll =)) rest)))
	(if appears
	  (lp (cdr l) accd (cons (car l) acci))
	  (lp (cdr l) (cons (car l) accd) acci))))))


(define (lset-union! = . rest)
  (apply lset-union = rest))		; XXX:optimize

(define (lset-intersection! = list1 . rest)
  (apply lset-intersection = list1 rest)) ; XXX:optimize

(define (lset-difference! = list1 . rest)
  (apply lset-difference = list1 rest))	; XXX:optimize

(define (lset-xor! = . rest)
  (apply lset-xor = rest))		; XXX:optimize

(define (lset-diff+intersection! = list1 . rest)
  (apply lset-diff+intersection = list1 rest)) ; XXX:optimize

;;; srfi-1.scm ends here
